/*
 * unimrcp_mc_client.c
 *
 *  Version 1.0
 *  Reviewed on: July 24, 2011 >>> Linux (CentOS 5.4 64 bit, Windows XP 32 bit and Windows Server 2008 64 bit)
 *    1) Porting to Windows
 *    2) Refactoring
 *    3) Adapting to UniMRCP 1.0.0
 *    4) Addition of comments
 *  Reviewer Simon Falardeau, Nuance Communications, Inc.
 *
 *  Created on: Nov 6, 2009 >>> Linux 
 *  Author: George M. Michelin, Nuance Communications, Inc.
 */

#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <apr.h>
#include <apr_getopt.h>
#include <apr_file_info.h>
#include <apr_pools.h>
#include <apr_thread_cond.h>
#include <apr_strings.h>
#include "unimrcp_client.h"
#include "mrcp_application.h"
#include "mrcp_client.h"
#include "mrcp_session.h"
#include "mrcp_message.h"
#include "mrcp_generic_header.h"
#include "mrcp_synth_header.h"
#include "mrcp_synth_resource.h"
#include "mrcp_recog_header.h"
#include "mrcp_recog_resource.h"
#include "apt_nlsml_doc.h"
#include "apt_log.h"
#include "apt_task.h"
#include "apt_task_msg.h"
#include "apt_consumer_task.h"
#include "apt_pool.h"
#include "apt_string.h"

//TYPES----------------------------------------------------------------------------------

typedef struct mc_app_channel_t mc_app_channel_t;
typedef struct mc_app_session_t mc_app_session_t;
typedef struct client_options_t client_options_t;
typedef struct mc_application_t mc_application_t;

//Structure holding MRCP applications with their name
struct mc_application_t
{
	const char			    * p_name;
	mrcp_application_t	* p_mrcp_app;
};

//Command line options
struct client_options_t 
{
	int				          num_apps;        //Number of applications to generate
	int			            num_iter;        //Number of iterations each application will run
	const char        * p_root_dir_path; //Root folder for outputfiles and for searching the config file
	apt_log_priority_e  log_priority;    //UniMRCP log priority
	apt_log_output_e    log_output;      //Output file for text logging
};

//Enumeration of the operations/roles that can have an application 
typedef enum 
{
	MC_SYNTH_DIGITS,  //Synthesize digits (each number)
	MC_RECOG_DIGITS,  //Recognize digits (each number)
	MC_SYNTH_NUMBER,  //Synthesize a number (number as a whole)
	MC_RECOG_NUMBER,  //Recognize a number (number as a whole)
	MC_NUM_ROLES      //Indicates the number of application types (4)
} mc_app_role_e;

/**
Declaration of the external object referenced by the channel.
This external object is attached to a UniMRCP channel in a 
transparent manner to be used in the callback methods. 
*/
struct mc_app_channel_t 
{
	FILE		    * p_synth_output_file;      //File in which to save received audio as a result of a SPEAK request
	char		    * p_recog_output_filename;  //File in which recognition results are stored in text format
	apt_bool_t	  streaming;                //True if the channel is open for sending or receiving audio
	apr_size_t	  time_to_complete;         //Time to fill a full audio frame when sending audio to the MRCP server
};

/**
Declaration of the external object referenced by the session.
This external object is attached to a UniMRCP session in a 
transparent manner to be used in the callback methods. 
*/
struct mc_app_session_t 
{
	char		     * p_number_string;   //Number used for synthesis or recognition
	int			       iteration;         //Current iteration in process
	mc_app_role_e  role;              //The role of the application
};

//Global variables ("s_" for static)-----------------------------------------------------

//Application control
int    s_num_apps_completed				= 0;  //Number of applications that completed all iterations
int    s_num_apps						      = 0;  //Number of applications to generate
int    s_num_app_iterations				= 0;  //Number if iterations each application must run
int    LOG_OUTPUT_CONSOLE_FILE    = 3;
int    MAX_NUMBER_STRING          = 9999; //Maximum value for the generated number string

//Profile to use in nuance.xml configuration file: drives the ports, the host and other 
//parameters used for MRCP, SIP and RTP communication
const char * s_nuanceProfileMRCPv2		= "nss2";  

//Pre-built content for building MRCP messages
const char   s_recog_digits_string[]	="<?xml version=\"1.0\"?><grammar xmlns=\"http://www.w3.org/2001/06/grammar\" xml:lang=\"en-US\" version=\"1.0\" mode=\"voice\" root=\"digits\"><rule id=\"digits\"><ruleref uri=\"builtin:grammar/digits?language=en-US\" /></rule></grammar>";
const char   s_recog_number_string[]	="<?xml version=\"1.0\"?><grammar xmlns=\"http://www.w3.org/2001/06/grammar\" xml:lang=\"en-US\" version=\"1.0\" mode=\"voice\" root=\"number\"><rule id=\"number\"><ruleref uri=\"builtin:grammar/number?language=en-US\" /></rule></grammar>";
const char   s_speak_digits_prefix[]	="<?xml version=\"1.0\"?><speak><say-as interpret-as=\"digits\">";
const char   s_speak_number_prefix[]	="<?xml version=\"1.0\"?><speak><say-as interpret-as=\"number\">";
const char   s_speak_suffix[]			    ="</say-as></speak>";

//APR variables (APR = ???) 
apr_pool_t		     * sp_app_pool		= NULL;   //APR pool for allocating memory
apr_thread_mutex_t * sp_recog_output_mutex;		//Mutex for writting in the recognition output file
apr_thread_mutex_t * sp_app_complete_mutex;		//Mutex for updating application process control variables
apr_thread_cond_t  * sp_app_complete_cond;		//Application thread completion signal

//Prototype declarations-----------------------------------------------------------------

//Utility methods
static const char *     mc_util_get_message_type	   (mrcp_message_type_e    a_type);
static const char *     mc_util_get_sig_status_str   (mrcp_sig_status_code_e a_sig_status);
static const apt_bool_t mc_util_extract_cmd_arguments(client_options_t     * ap_options, 
													                            int				             ap_argc, 
													                            const char * const   * ap_argv);

//Application methods
static apt_bool_t mc_application_initialize	 (mrcp_application_t       * ap_application);
static apt_bool_t mc_application_handler     (const mrcp_app_message_t * ap_app_message);

//Client methods
static apt_bool_t mc_client_application_start(mrcp_client_t      * ap_client);      //Client + application method
static apt_bool_t mc_client_init_thread_ctrl ();
static apt_bool_t mc_client_init_apr	       ();
static apt_bool_t mc_client_stop			       (mrcp_client_t		   * ap_client);
static apt_bool_t mc_client_load_options     (client_options_t   * ap_options, 
										                          int				           ap_argc, 
										                          const char * const * ap_argv);
static mrcp_client_t *  mc_client_init	     (int 			           ap_argv, 
										                          const char * const * ap_argvc  );

//Callback methods for MRCP event handling
static apt_bool_t mc_application_on_session_update   (mrcp_application_t	  * ap_application, 
                        														  mrcp_session_t			  * ap_session, 
														                          mrcp_sig_status_code_e  a_status);

static apt_bool_t mc_application_on_session_terminate(mrcp_application_t		* ap_application, 
														                          mrcp_session_t		  	* ap_session, 
														                          mrcp_sig_status_code_e  a_status);

static apt_bool_t mc_application_on_channel_add      (mrcp_application_t		* ap_application, 
													                          	mrcp_session_t			  * ap_session, 
														                          mrcp_channel_t			  * ap_channel, 
														                          mrcp_sig_status_code_e  a_status);

static apt_bool_t mc_application_on_channel_remove   (mrcp_application_t		* ap_application, 
														                          mrcp_session_t		  	* ap_session, 
														                          mrcp_channel_t		  	* ap_channel, 
														                          mrcp_sig_status_code_e  a_status);

static apt_bool_t mc_application_on_message_receive  (mrcp_application_t		* ap_application, 
														                          mrcp_session_t	 		  * ap_session, 
														                          mrcp_channel_t	 		  * ap_channel, 
														                          mrcp_message_t			  * ap_message);

//Callback methods for synthesizer RTP stream (outbound)
static apt_bool_t synth_stream_destroy(mpf_audio_stream_t * ap_stream);
static apt_bool_t synth_stream_open   (mpf_audio_stream_t * ap_stream, mpf_codec_t * ap_codec);
static apt_bool_t synth_stream_close  (mpf_audio_stream_t * ap_stream);
static apt_bool_t synth_stream_write  (mpf_audio_stream_t * ap_stream, const mpf_frame_t * ap_frame);

//Callback methods for "recognizer" RTP stream (inbound)
static apt_bool_t recog_stream_destroy(mpf_audio_stream_t * ap_stream);
static apt_bool_t recog_stream_open   (mpf_audio_stream_t * ap_stream, mpf_codec_t * ap_codec);
static apt_bool_t recog_stream_close  (mpf_audio_stream_t * ap_stream);
static apt_bool_t recog_stream_read   (mpf_audio_stream_t * ap_stream, mpf_frame_t * ap_frame);

//Callback methods for MRCP message processing
static mrcp_message_t * app_speak_message_create         (mrcp_session_t * ap_session, 
														                              mrcp_channel_t * ap_channel);
static mrcp_message_t * app_define_grammar_message_create(mrcp_session_t * ap_session, 
														                              mrcp_channel_t * ap_channel);
static mrcp_message_t * app_recognize_message_create     (mrcp_session_t * ap_session, 
														                              mrcp_channel_t * ap_channel);
static char *		        mc_application_parse_result      (mrcp_message_t * ap_message);

//Callback method structures used by UniMRCP framework-----------------------------------

/**
Application callback methods.  
Linked to applications in UniMRCP framework and called for 
message handling, session handling and channel control by UniMRCP 
*/
static const mrcp_app_message_dispatcher_t mc_application_dispatcher = 
{
	mc_application_on_session_update,
	mc_application_on_session_terminate,
	mc_application_on_channel_add,
	mc_application_on_channel_remove,
	mc_application_on_message_receive
};

/**
Synthesizer stream callback methods (outbound) for receiving audio data from the MRCP server (RTP channel). 
The complete structure contains inbound and outbound callback methods: NULL is set for inbound. 
- Destroy
- Inbound  open  
- Inbound  close
- Inbound  read
- Outbound open
- Outbound close
- Outbound read
*/
static const mpf_audio_stream_vtable_t synth_stream_vtable =
{
	synth_stream_destroy,
	NULL,
	NULL,
	NULL,
	synth_stream_open,
	synth_stream_close,
	synth_stream_write
};

/**
Recognizer stream callback methods (inbound) for sending audio data to the MRCP server (RTP channel).
The complete structure contains inbound and outbound call back methods: NULL is set for outbound. 
- Destroy
- Inbound  open  
- Inbound  close
- Inbound  read
- Outbound open
- Outbound close
- Outbound read
*/
static const mpf_audio_stream_vtable_t recog_stream_vtable =
{
	recog_stream_destroy,
	recog_stream_open,
	recog_stream_close,
	recog_stream_read,
	NULL,
	NULL,
	NULL
};



// Client application methods ----------------------------------------------------------------

/**
Returns the string version of MRCP message type enumeration values.

@PARAMS:
  - a_type : MRCP message type from UniMRCP (MRCP_MESSAGE_TYPE_*)
  - return : string replica of the enumeration item name
*/
static const char * mc_util_get_message_type_str( mrcp_message_type_e a_type)
{
	char * vp_typeStr;

	switch (a_type)
	{
		case MRCP_MESSAGE_TYPE_UNKNOWN:
			vp_typeStr = "MRCP_MESSAGE_TYPE_UNKNOWN";
			break;
		case MRCP_MESSAGE_TYPE_REQUEST:
			vp_typeStr = "MRCP_MESSAGE_TYPE_REQUEST";
			break;
		case MRCP_MESSAGE_TYPE_RESPONSE:
			vp_typeStr = "MRCP_MESSAGE_TYPE_RESPONSE";
			break;
		case MRCP_MESSAGE_TYPE_EVENT:
			vp_typeStr = "MRCP_MESSAGE_TYPE_EVENT";
			break;
		default:
			vp_typeStr = "UNKNOWN_TYPE";
	}
	return vp_typeStr;
}

/**
Returns the string version of "MRCP signal status" type enumeration values.

@PARAMS:
  - a_sig_status: MRCP signal status type from UniMRCP (MRCP_SIG_STATUS_CODE_*)
  - return      : string replica of the enumeration item name
*/
static const char * mc_util_get_sig_status_str( mrcp_sig_status_code_e a_sig_status)
{
	char * vp_typeStr;

	switch (a_sig_status)
	{
		case MRCP_SIG_STATUS_CODE_SUCCESS:
			vp_typeStr = "MRCP_SIG_STATUS_CODE_SUCCESS";
			break;
		case MRCP_SIG_STATUS_CODE_FAILURE:
			vp_typeStr = "MRCP_SIG_STATUS_CODE_FAILURE";
			break;
		case MRCP_SIG_STATUS_CODE_TERMINATE:
			vp_typeStr = "MRCP_SIG_STATUS_CODE_TERMINATE";
			break;
		case MRCP_SIG_STATUS_CODE_CANCEL:
			vp_typeStr = "MRCP_SIG_STATUS_CODE_CANCEL";
			break;
		default:
			vp_typeStr = "UNKNOWN_TYPE";
	}
	return vp_typeStr;
}

/** 
Creates a session instance linked to an application and to the profile defined globally
under variable "s_nuanceProfileMRCPv2" and creates the "external" session object used in 
the callback methods.

A session is bound to an application which comprises a set of recognition requests, 
including define-grammar, and speak requests.
@PARAMS
  - ap_application   : UniMRCP application driving the recognition and speak requests
  - a_role			     : used to guide in the generation of requests
  - a_iteration      : number of time the application will repeat the combination of requests
  - ap_number_string : list of digits/numbers to recognize and/or to speak
  - return			     : an MRCP session linked to the given application 
*/
static mrcp_session_t* mc_application_session_create(mrcp_application_t * ap_application, 
													                           mc_app_role_e		    a_role, 
													                           int				          a_iteration, 
													                           const char			    * ap_number_string)
{
	mrcp_session_t   * vp_session; 

	//External object attached to the session: used from different callback methods 
	mc_app_session_t * vp_mc_app_session	= apr_palloc(sp_app_pool, sizeof(mc_app_session_t));
	vp_mc_app_session->role					      = a_role;
	vp_mc_app_session->iteration			    = a_iteration;
	vp_mc_app_session->p_number_string		= apr_pstrdup(sp_app_pool, ap_number_string);

	//UniMRCP session creation
	vp_session = mrcp_application_session_create(ap_application,
	   										                       s_nuanceProfileMRCPv2,
											                         vp_mc_app_session);
	return vp_session;
}

/** 
Creates a channel used to tunnel, between the MRCP client and server, either audio input 
to be recognized or audio output to obtain from the server as a result of a SPEAK request.

The creation of a channel instance in UniMRCP is done by calling method 
mrcp_application_channel_create(...).

In MRCP, channel creation is done from SIP INVITE message which corresponds to
the request to start a session.  UniMRCP will use the attributes of the channel instance(s)
to populate the SIP INVITE message.

For example, the following lines from a SIP INVITE are for the creation of a channel:
"...
m=audio 49170 RTP/AVP 0		> Protocole and port 
a=rtpmap:0 pcmu/8000			> Derived from UniMRCP codec: audio encoding used in the channel
a=recvonly						    > Data flow direction: input, output or both
a=mid:1							      > Channel ID
..."

@PARAMS:
	- ap_session: the session for which the channel will be created 
	- return    : a channel ready for audio streaming
*/
static mrcp_channel_t* mc_application_channel_create(mrcp_session_t * ap_session)
{
	mrcp_channel_t			    * vp_channel			    = NULL;
	mpf_termination_t   		* vp_termination		  = NULL;
	mpf_codec_descriptor_t  * vp_codec_descriptor	= NULL;
	apr_size_t				        v_mrcp_resource		  = 0;

	//External object attached previously to the session will be used to create the channel
	mc_app_session_t        * vp_mc_session	     	= mrcp_application_session_object_get(ap_session);

	//External object attached to the channel: used from different callback methods
	mc_app_channel_t        * vp_mc_app_channel   = apr_palloc(ap_session->pool,sizeof(mc_app_channel_t));
	vp_mc_app_channel->p_synth_output_file			  = NULL;
	vp_mc_app_channel->p_recog_output_filename	  = NULL;
	vp_mc_app_channel->streaming				      	  = FALSE;
	vp_mc_app_channel->time_to_complete				    = 0;

	//Determine if we have an outbound (send / SPEAK requests) from the server perspective
	//or an inbound (receive / RECOGNITION requests) and set the termination context accordingly. 
	if ( vp_mc_session->role == MC_SYNTH_DIGITS || vp_mc_session->role == MC_SYNTH_NUMBER) 
	{
		v_mrcp_resource = MRCP_SYNTHESIZER_RESOURCE;
		vp_termination  = mrcp_application_sink_termination_create( ap_session,           /* session to which the termination belongs to */
						 										                               &synth_stream_vtable,	/* virtual methods table of audio stream */
																                                vp_codec_descriptor,  /* codec descriptor of audio stream (NULL by default) */
																                                vp_mc_app_channel);   /* object to associate */
	}
	else 
	{
		v_mrcp_resource  = MRCP_RECOGNIZER_RESOURCE;
		vp_termination	 = mrcp_application_source_termination_create( ap_session,          /* session to which the termination belongs to */
													  			                                &recog_stream_vtable, /* virtual methods table of audio stream */
																                                   vp_codec_descriptor, /* codec descriptor of audio stream (NULL by default) */
																                                   vp_mc_app_channel);  /* object to associate */
	}

	//Create and return the channel
	vp_channel = mrcp_application_channel_create( ap_session,         /* session, channel belongs to */
												                        v_mrcp_resource,    /* MRCP resource identifier */
												                        vp_termination,     /* media termination, used to terminate audio stream */
												                        NULL,               /* RTP descriptor, used to create RTP termination (NULL by default) */
												                        vp_mc_app_channel); /* object to associate */
	return vp_channel;
}

/** 
Creates a session and adds an output channel to it to process SPEAK requests and
to receive the audio output resulting from it.  
Also creates an external application object used in callback methods. 

In more details:
1) Generates a random number used in the SPEAK requests and stores this in the variable "number_string" 
2) Creates an external application object, that contains the number string, 
   and attaches it to the UniMRCP application
3) Creates a UniMRCP session instance
4) Creates a UniMRCP channel and attaches it to the session

As a result, the application is ready to launch and SPEAK request messages are buffered in UniMRCP.
In this example client, when the session ends a new one is started for the same application
from the session terminate callback method but with a different role.

@PARAMS:
 - ap_application: UniMRCP application for which a session and a channel are created.
 - return        : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_application_initialize(mrcp_application_t * ap_application)
{
	int				        v_random_number		    = rand() % MAX_NUMBER_STRING;
	char			      * vp_number_string		  = apr_itoa(sp_app_pool, v_random_number);
	mrcp_session_t	* vp_bootstrap_session	= NULL;
	mrcp_channel_t	* vp_bootstrap_channel	= NULL;

	//Create a session for an application that makes speak requests for the digits in "number_string".
	vp_bootstrap_session = mc_application_session_create(ap_application, MC_SYNTH_DIGITS, 1, vp_number_string);
	if(!vp_bootstrap_session) 
	{
		return FALSE;
	}

	//Create a channel for a "speak request" session
	vp_bootstrap_channel = mc_application_channel_create(vp_bootstrap_session);
	if(!vp_bootstrap_channel) 
	{
		mrcp_application_session_destroy(vp_bootstrap_session);
		return FALSE;
	}

	//Add the channel to the session: will result in calling back "mc_application_on_channel_add" which
	//will create SPEAK requests in UniMRCP.  
	if(mrcp_application_channel_add(vp_bootstrap_session, vp_bootstrap_channel) != TRUE) 
	{
		//The channel could not be added, the session is not usable.
		mrcp_application_session_destroy(vp_bootstrap_session);
		return FALSE;
	}

	return TRUE;
}

//Callback methods -------------------------------------------------------

/** 
Called by the UniMRCP framework to obtain callback methods of an application.
This handler method is linked to the application from the "MRCP application create" method. 

@PARAMS:
 - ap_app_message: contains application, session, channel information and signaling or control message
 - return        : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_application_handler(const mrcp_app_message_t * ap_app_message)
{
	return mrcp_application_message_dispatch(&mc_application_dispatcher, ap_app_message);
}

/** 
Called from UniMRCP framework when a session is updated.
Nothing is done in this example outside the normal UniMRCP process but it would be 
possible to react.

@PARAMS:
 - application : UniMRCP application driving the recognition and speak requests
 - ap_session	 : UniMRCP session being updated 
 - a_status		 : The status of the session update process on UniMRCP side
 - return      : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_application_on_session_update(mrcp_application_t   * ap_application, 
                                                   mrcp_session_t	      * ap_session, 
                                                   mrcp_sig_status_code_e a_status)
{
	return TRUE;
}

/** 
Called by the UniMRCP framework when a session is terminated.
Creates a new session and channel using a new role in order to test each role.
It would be possible to use the same session: creating a new session was decided 
without prejudice to the alternative.

The first time this method is called, the role of the "session" is MC_SYNTH_DIGITS.
Then, when the session terminates, a new session is created using the next role from 
"mc_app_role_e":
- MC_RECOG_DIGITS,
- MC_SYNTH_NUMBER,
- MC_RECOG_NUMBER,
- MC_NUM_ROLES >>> not used.

Also, the concept of iteration is applied. When session termination is called 
and the current role is MC_RECOG_NUMBER (the last role in the enumeration):
- If we have not reached the number of defined iterations 
	> the role is reset to MC_SYNTH_DIGITS
	> Number of iterations is incremented
- If this was the last iteration, the application "stops" and a signal is sent to 
  apr thread handling accordingly.

@PARAMS:
 - application : UniMRCP application driving the recognition and speak requests
 - ap_session	 : UniMRCP session being terminated
 - a_status		 : The status of the session termination process on UniMRCP side
 - return      : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_application_on_session_terminate(mrcp_application_t  	* ap_application, 
                                                      mrcp_session_t	  	  * ap_session, 
                                                      mrcp_sig_status_code_e  a_status)
{
	mrcp_session_t   * vp_new_session = NULL;
	mrcp_channel_t   * vp_new_channel = NULL;
	mc_app_session_t * vp_mc_session  = mrcp_application_session_object_get(ap_session);
	mrcp_application_session_destroy(ap_session);

	if (vp_mc_session->role < MC_RECOG_NUMBER) 
	{
		//Increment role, create new session and channel using the same number string
		vp_new_session = mc_application_session_create( ap_application, 
														                        vp_mc_session->role + 1, 
														                        vp_mc_session->iteration, 
														                        vp_mc_session->p_number_string);
		if(!vp_new_session) 
		{
			return FALSE;
		}
		//Create and add the channel
		vp_new_channel = mc_application_channel_create(vp_new_session);
		if(!vp_new_channel) 
		{
			mrcp_application_session_destroy(vp_new_session);
			return FALSE;
		}

		//Add the channel to the session: will result in calling back "mc_application_on_channel_add" which
		//will create MRCP requests in UniMRCP. 
		if (mrcp_application_channel_add(vp_new_session, vp_new_channel) != TRUE)	
		{
			mrcp_application_session_destroy(vp_new_session);
			return FALSE;
		}
	}
	else //(vp_mc_session->role >= MC_RECOG_NUMBER)
	{
		//Last role but we did not reach the number of desired iterations
		if (vp_mc_session->iteration < s_num_app_iterations) 
		{
			//Increment iteration, reset role to MC_SYNTH_DIGITS, assign new random number, create new session and channel
			int    v_random_number	= rand() % MAX_NUMBER_STRING;
			char * vp_number_string = apr_itoa(sp_app_pool, v_random_number);

			vp_new_session = mc_application_session_create( ap_application, 
															                        MC_SYNTH_DIGITS, 
														                          vp_mc_session->iteration + 1,  
														                          vp_number_string);
			if(!vp_new_session) 
			{
				return FALSE;
			}

			vp_new_channel = mc_application_channel_create(vp_new_session);
			if(!vp_new_channel) 
			{
				mrcp_application_session_destroy(vp_new_session);
				return FALSE;
			}

			//Add the channel to the session: will result in calling back "mc_application_on_channel_add" which
			//will create SPEAK requests in UniMRCP. 
			if (mrcp_application_channel_add(vp_new_session, vp_new_channel) != TRUE)	
			{
				mrcp_application_session_destroy(vp_new_session);
				return FALSE;
			}
		}
		else //(vp_mc_session->iteration => s_num_app_iterations)
		{
			//Last role in sequence is MC_RECOG_NUMBER and desired number of iterations completed
			//Send "application complete" signal and adjust application tracking counter
			apr_thread_mutex_lock(sp_app_complete_mutex);
			s_num_apps_completed=s_num_apps_completed + 1;
			apr_thread_cond_signal(sp_app_complete_cond);
			apr_thread_mutex_unlock(sp_app_complete_mutex);
		} //end if (vp_mc_session->iteration < s_num_app_iterations)
	} //end if (vp_mc_session->role < MC_RECOG_NUMBER)
	return TRUE;
}

/** 
Called by UniMRCP framework when a channel has been added.  
If the channel was added successfully, an MRCP message is created and sent:
- If the session role is "synthesize", a SPEAK request is sent
- if the session role is "recognize", a DEFINE-GRAMMAR message is sent.
  The RECOGNIZE message is sent from "mc_application_on_define_grammar" if DEFINE-GRAMMAR
  was successfully completed.  Note that grammars can be statically defaulted in NSS and can
  be defined at runtime as done in this example.

Also opens files to capture speech audio and recognition results.

@PARAMS:
 - application : UniMRCP application driving the recognition and speak requests
 - ap_session	 : UniMRCP session used by the application 
 - ap_channel	 : UniMRCP channel being added to the session
 - a_status		 : The status of the channel addition process on UniMRCP side
 - return      : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_application_on_channel_add(mrcp_application_t		* ap_application, 
							                                  mrcp_session_t			  * ap_session, 
                                                mrcp_channel_t			  * ap_channel, 
                                                mrcp_sig_status_code_e  a_status)
{
	char                   * vp_synth_file_name	= NULL;
	char                   * vp_recog_file_name	= NULL;
	char                   * vp_file_path       = NULL;
	char                   * vp_synth_file_path	= NULL;
	const apt_dir_layout_t * vp_dir_layout      = mrcp_application_dir_layout_get(ap_application);
	mc_app_session_t	     * vp_mc_session      = mrcp_application_session_object_get(ap_session);
	mc_app_channel_t	     * vp_mc_channel      = mrcp_application_channel_object_get(ap_channel);
	mrcp_message_t		     * vp_mrcp_message    = NULL;	

	if(a_status == MRCP_SIG_STATUS_CODE_SUCCESS) 
	{
		//DIGITS
		if (vp_mc_session->role == MC_SYNTH_DIGITS || vp_mc_session->role == MC_RECOG_DIGITS) 
		{
			vp_synth_file_name = apr_pstrcat(ap_session->pool, "synth-digits-", vp_mc_session->p_number_string, ".pcm", NULL);
			vp_recog_file_name = apr_pstrcat(ap_session->pool, "recog-digits-", vp_mc_session->p_number_string, ".txt", NULL);
		}
		//NUMBERS
		else if (vp_mc_session->role == MC_SYNTH_NUMBER || vp_mc_session->role == MC_RECOG_NUMBER) 
		{
			vp_synth_file_name = apr_pstrcat(ap_session->pool, "synth-number-", vp_mc_session->p_number_string, ".pcm", NULL);
			vp_recog_file_name = apr_pstrcat(ap_session->pool, "recog-number-", vp_mc_session->p_number_string, ".txt", NULL);
		}

		if (vp_mc_session->role == MC_SYNTH_DIGITS || vp_mc_session->role == MC_SYNTH_NUMBER) 
		{
			//Create and send a SPEAK request
			vp_mrcp_message = app_speak_message_create(ap_session, ap_channel);
			if(vp_mrcp_message) 
			{
				mrcp_application_message_send(ap_session, ap_channel, vp_mrcp_message);
			}

      //Open the file used to capture audio resulting from SPEAK requests
			vp_file_path = apt_datadir_filepath_get(vp_dir_layout, vp_synth_file_name, ap_session->pool);
			if(vp_file_path) 
			{
				vp_mc_channel->p_synth_output_file = fopen(vp_file_path, "wb");

				if(vp_mc_channel->p_synth_output_file) 
				{
					apt_log(APT_LOG_MARK, APT_PRIO_INFO, "Set [%s] as Speech Destination", vp_file_path);
				}
				else 
				{
					apt_log(APT_LOG_MARK, APT_PRIO_INFO, "Cannot Open [%s]", vp_file_path);
				}
			}
		}
		else if (vp_mc_session->role == MC_RECOG_DIGITS || vp_mc_session->role == MC_RECOG_NUMBER) 
		{
			//Create and send DEFINE GRAMMAR 
			vp_mrcp_message = app_define_grammar_message_create(ap_session, ap_channel);
			if(vp_mrcp_message) 
			{
				mrcp_application_message_send(ap_session, ap_channel, vp_mrcp_message);
			}

			//Set the source file for generating speech to be recognized using the audio captured from the SPEAK requests
			vp_synth_file_path	= apt_datadir_filepath_get(vp_dir_layout, vp_synth_file_name, ap_session->pool);
			if (vp_synth_file_path) 
			{
				vp_mc_channel->p_synth_output_file = fopen(vp_synth_file_path, "rb");

				if(vp_mc_channel->p_synth_output_file) 
				{
					apt_log(APT_LOG_MARK, APT_PRIO_INFO, "Set [%s] as Speech Source", vp_synth_file_path);
				}
				else 
				{
					apt_log(APT_LOG_MARK, APT_PRIO_INFO, "Cannot Find [%s]", vp_synth_file_path);
					vp_mc_channel->time_to_complete = 15000;
				}
			}
			//File for storing the recognition result
			vp_mc_channel->p_recog_output_filename = apt_datadir_filepath_get(vp_dir_layout, vp_recog_file_name, ap_session->pool);
			apt_log(APT_LOG_MARK, APT_PRIO_INFO, "Set [%s] as Recognition Results Destination", vp_mc_channel->p_recog_output_filename);
		}
	}
	else 
	{
		//Channel could not be created, just terminate the demo
		mrcp_application_session_terminate(ap_session);
	}
	return TRUE;
}

/** 
Called by UniMRCP framework when a channel has been removed.
Closes the file used to capture output from the server (SPEAK request audio stream) and
calls for the termination of the session.

@PARAMS:
 - application : UniMRCP application driving the recognition and speak requests
 - ap_session	 : UniMRCP session used by the application
 - ap_channel	 : UniMRCP channel being removed from the session
 - a_status		 : The status of the channel removal process on UniMRCP side
 - return      : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_application_on_channel_remove(mrcp_application_t	    * ap_application, 
                                                   mrcp_session_t		      * ap_session, 
                                                   mrcp_channel_t         * ap_channel, 
                                                   mrcp_sig_status_code_e   a_status)
{
	
	mc_app_channel_t * vp_mc_channel  = mrcp_application_channel_object_get(ap_channel);
	FILE			 * vp_synth_output_file = vp_mc_channel->p_synth_output_file;
	
	if(vp_synth_output_file) 
	{
		vp_mc_channel->p_synth_output_file = NULL;
		fclose(vp_synth_output_file);
	}
	
	return mrcp_application_session_terminate(ap_session);
}

/** 
Called when the message DEFINE-GRAMMAR was received and confirmed by the server.
Creates and sends a recognition request.

@PARAMS
 - application : UniMRCP application requesting the grammar definition
 - ap_session	 : UniMRCP session used by the application
 - ap_channel	 : UniMRCP channel for which the grammar was defined 
*/
static apt_bool_t mc_application_on_define_grammar( mrcp_application_t * ap_application, 
                                                    mrcp_session_t		 * ap_session, 
                                                    mrcp_channel_t		 * ap_channel)
{
	mrcp_message_t         * vp_mrcp_message;
	const apt_dir_layout_t * vp_dir_layout = mrcp_application_dir_layout_get(ap_application);

	// create and send RECOGNIZE request
	vp_mrcp_message = app_recognize_message_create(ap_session, ap_channel);
	if(vp_mrcp_message) 
	{
		mrcp_application_message_send(ap_session, ap_channel, vp_mrcp_message);
	}

	return TRUE;
}

/** 
Called by the UniMRCP framework when a message, or a confirmation following a request,
is received from the MRCP server.  Reacts to the message or confirmation with the next "step".
The validity of the messages is bound to the role of the application.
- SYNTHESIZER_SPEAK_COMPLETE      : received the audio, close the channel
- MRCP_REQUEST_STATE_INPROGRESS   : receiving audio, do nothing, wait for the SPEAK COMPLET message
- RECOGNIZER_DEFINE_GRAMMAR       : Grammar define completed, send a recognition request
- RECOGNIZER_RECOGNIZE            : Recognize request accepted, turn on streaming
- RECOGNIZER_RECOGNITION_COMPLETE : Extract the recognition result and save to file
- RECOGNIZER_START_OF_INPUT       : Input started to be received on server side, nothing to do in our case

@PARAMS:
 - application : UniMRCP application to which the message belongs
 - ap_session  : UniMRCP session used by the application
 - ap_channel	 : UniMRCP channel for the audio stream
 - ap_message	 : UniMRCP message received as a result of a request sent out by the application
 - return      : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_application_on_message_receive(mrcp_application_t * ap_application, 
                                                    mrcp_session_t     * ap_session, 
                                                    mrcp_channel_t     * ap_channel, 
                                                    mrcp_message_t     * ap_message)
{
	mc_app_session_t * vp_mc_session = mrcp_application_session_object_get(ap_session);
	mc_app_channel_t * vp_mc_channel = mrcp_application_channel_object_get(ap_channel);

	if (vp_mc_session->role == MC_SYNTH_DIGITS || vp_mc_session->role == MC_SYNTH_NUMBER) 
	{
		//ROLE is "synthesize" so we are dealing with SPEAK request responses and events 
		if(ap_message->start_line.message_type == MRCP_MESSAGE_TYPE_RESPONSE) 
		{
			//Received MRCP response related to SPEAK request
			if(ap_message->start_line.method_id == SYNTHESIZER_SPEAK) 
			{
				//Received the response to SPEAK request
				if(ap_message->start_line.request_state == MRCP_REQUEST_STATE_INPROGRESS) 
				{
					//Waiting for SPEAK-COMPLETE event, do nothing.
				}
				else 
				{
					//Received unexpected response, remove channel
					mrcp_application_channel_remove(ap_session, ap_channel);
				}
			}
			else 
			{
				//Received unexpected response
				//Should the channel be removed?
			}
		}
		else if(ap_message->start_line.message_type == MRCP_MESSAGE_TYPE_EVENT) 
		{
			//Received MRCP event related to SPEAK request
			if(ap_message->start_line.method_id == SYNTHESIZER_SPEAK_COMPLETE) 
			{
				//Received SPEAK-COMPLETE event, remove channel since there is no more audio to receive
				mrcp_application_channel_remove(ap_session, ap_channel);
			}
		}
	}
	else if (vp_mc_session->role == MC_RECOG_DIGITS || vp_mc_session->role == MC_RECOG_NUMBER) 
	{
		//Role is "recognize" so we are dealing with GRAMMAR and RECOGNIZE request responses and events
		if(ap_message->start_line.message_type == MRCP_MESSAGE_TYPE_RESPONSE) 
		{
			//Received MRCP response related to DEFINE-GRAMMAR or RECOGNIZE request
			if(ap_message->start_line.method_id == RECOGNIZER_DEFINE_GRAMMAR) 
			{
				//Received the response to DEFINE-GRAMMAR request
				if(ap_message->start_line.request_state == MRCP_REQUEST_STATE_COMPLETE) 
				{
					mc_application_on_define_grammar(ap_application, ap_session, ap_channel);
				}
				else 
				{
					//Received unexpected response, remove channel
					mrcp_application_channel_remove(ap_session, ap_channel);
				}
			}
			else if(ap_message->start_line.method_id == RECOGNIZER_RECOGNIZE) 
			{
				//Received the response to DEFINE-GRAMMAR or RECOGNIZE request
				if(ap_message->start_line.request_state == MRCP_REQUEST_STATE_INPROGRESS) 
				{
          //Recognition in progress, audio data can be sent to the MRCP server
					//Turn stream on -> will impact next call to recog_stream_read
					vp_mc_channel->streaming = TRUE;
				}
				else 
				{
					//Received unexpected response, remove channel
					mrcp_application_channel_remove(ap_session, ap_channel);
				}
			}
			else 
			{
				// received unexpected response
			}
		}
		else if(ap_message->start_line.message_type == MRCP_MESSAGE_TYPE_EVENT) 
		{
			//Received MRCP event related to RECOGNIZE request
			if(ap_message->start_line.method_id == RECOGNIZER_RECOGNITION_COMPLETE) 
			{
				//Extract recognition result and write it to the recognition output file
				char * vp_parse_result		= mc_application_parse_result(ap_message);
				FILE * vp_recog_output_file = NULL; 
				
				//When the result is obtained, add it to the recognition output file.
				if (vp_parse_result) 
				{		
					apr_thread_mutex_lock(sp_recog_output_mutex);
					vp_recog_output_file = fopen(vp_mc_channel->p_recog_output_filename, "wb");
					fwrite(vp_parse_result, strlen(vp_parse_result), 1, vp_recog_output_file);
					fclose (vp_recog_output_file);
					apr_thread_mutex_unlock(sp_recog_output_mutex);
				}

				//No more streaming required, turn it off -> will impact next call to recog_stream_read
				vp_mc_channel->streaming = FALSE;
				mrcp_application_channel_remove(ap_session,ap_channel);
			}
			else if(ap_message->start_line.method_id == RECOGNIZER_START_OF_INPUT) 
			{
				//Received start-of-input, do whatever you need here
			}
		}
	}

	return TRUE;
}

//MRCP message creation methods-------------------------------------------------------

/** 
Create and return an MRCP SPEAK request message.  The generic header is generated by the UniMRCP framework
and some additions are done in this method: 
- content type    = application/ssml+xml 
- voice age value = 28
- SSML with text to synthesize.

SPEAK request example:
MRCP/2.0 489 SPEAK 543257					
Channel-Identifier:32AECB23433802@speechsynth
Voice-Age:28                               >>>Added in this method (from here and below)
Content-Type:application/ssml+xml		
Content-Length:...							           >>>Length value generated by UniMRCP framework

<?xml version=\"1.0\"?>
<speak><say-as interpret-as=\"digits\">"	 >>>Either digits or numbers depending on the role 
	<number_string>							             >>>Our generated sequence of digits	
</say-as></speak>"

@PARAMS:
 - ap_session : UniMRCP session relevant for the message
 - ap_channel : UniMRCP channel through which audio stream will be received
 - return     : a fully qualified UniMRCP SPEAK request message
*/
static mrcp_message_t* app_speak_message_create(mrcp_session_t * ap_session, 
                                                mrcp_channel_t * ap_channel)
{
	mc_app_session_t * vp_mc_session   = mrcp_application_session_object_get(ap_session);
	char             * vp_speak_string = NULL;

	/* create MRCP message */
	mrcp_message_t * vp_mrcp_message = mrcp_application_message_create(ap_session,ap_channel, SYNTHESIZER_SPEAK);
	if(vp_mrcp_message) 
	{
		mrcp_generic_header_t * vp_generic_header;
		mrcp_synth_header_t   * vp_synth_header;

		//Add "content type = application/synthesis+ssml" to the generic header
		vp_generic_header = mrcp_generic_header_prepare(vp_mrcp_message);
		if(vp_generic_header) 
		{
			apt_string_assign(&vp_generic_header->content_type, "application/synthesis+ssml", vp_mrcp_message->pool);
			mrcp_generic_header_property_add(vp_mrcp_message, GENERIC_HEADER_CONTENT_TYPE);
		}

		//Add "voice_param.age = 28" to the SPEAK request header section
		vp_synth_header = mrcp_resource_header_prepare(vp_mrcp_message);
		if(vp_synth_header) 
		{
			vp_synth_header->voice_param.age = 28;
			mrcp_resource_header_property_add(vp_mrcp_message, SYNTHESIZER_HEADER_VOICE_AGE);
		}

		//Add the content "ssml" with the request text to synthesize and the request format
		if (vp_mc_session->role == MC_SYNTH_DIGITS) 
		{
			vp_speak_string = apr_pstrcat(vp_mrcp_message->pool, s_speak_digits_prefix, vp_mc_session->p_number_string, s_speak_suffix, NULL);
		}
		else if (vp_mc_session->role == MC_SYNTH_NUMBER) 
		{
			vp_speak_string = apr_pstrcat(vp_mrcp_message->pool, s_speak_number_prefix, vp_mc_session->p_number_string, s_speak_suffix, NULL);
		}
		apt_string_assign_n(&vp_mrcp_message->body, vp_speak_string, strlen(vp_speak_string), vp_mrcp_message->pool);
	}
	return vp_mrcp_message;
}

/** 
Create and returns an MRCP DEFINE-GRAMMAR request message.  
The generic header is generated by the UniMRCP framework and some additions are done in this method:
- "content type = application/srgs+xml"
- "content is   = request1@form-level.store"
- Grammar definition xml content

Example:
MRCP/2.0 589 DEFINE-GRAMMAR 543257
Channel-Identifier:32AECB23433801@speechrecog
Content-Type:application/srgs+xml					>>>Added in this method (from here and below) 
Content-ID:<request1@form-level.store>
Content-Length:...									      >>>Length value generated by UniMRCP framework

"<?xml version=\"1.0\"?>						    	>>>Content
<grammar xmlns=\"http://www.w3.org/2001/06/grammar\" 
         xml:lang=\"en-US\" 
		 version=\"1.0\" 
		 mode=\"voice\" 
		 root=\"digits\">
	<rule id=\"digits\">
		<ruleref uri=\"builtin:grammar/digits?language=en-US\" />  >>>Either digits or numbers depending on the role
	</rule>
</grammar>";

@PARAMS:
 - ap_session : UniMRCP session relevant for the message
 - ap_channel : UniMRCP channel through which audio stream will be sent for recognition 
	              based on this grammar
 - return	    : a fully qualified UniMRCP DEFINE-GRAMMAR request message
*/
static mrcp_message_t* app_define_grammar_message_create(mrcp_session_t * ap_session,
                                                         mrcp_channel_t * ap_channel)
{
	mc_app_session_t * vp_mc_session   = mrcp_application_session_object_get(ap_session);
	mrcp_message_t	 * vp_mrcp_message = mrcp_application_message_create(ap_session,ap_channel, RECOGNIZER_DEFINE_GRAMMAR);

	if(vp_mrcp_message) 
	{
		mrcp_generic_header_t * vp_generic_header;

		//Add "content type = application/srgs+xml" and "Content-ID:<request1@form-level.store>" to the generic header
		vp_generic_header = mrcp_generic_header_prepare(vp_mrcp_message);
		if(vp_generic_header) 
		{
			if(vp_mrcp_message->start_line.version == MRCP_VERSION_2) 
			{
				apt_string_assign(&vp_generic_header->content_type, "application/srgs+xml", vp_mrcp_message->pool);
			}
			else
			{
				apt_string_assign(&vp_generic_header->content_type, "application/grammar+xml", vp_mrcp_message->pool);
			}
			mrcp_generic_header_property_add(vp_mrcp_message, GENERIC_HEADER_CONTENT_TYPE);
			apt_string_assign(&vp_generic_header->content_id, "request1@form-level.store", vp_mrcp_message->pool);
			mrcp_generic_header_property_add(vp_mrcp_message, GENERIC_HEADER_CONTENT_ID);
		}

		//Adding the Grammar definition xml body
		if (vp_mc_session->role == MC_RECOG_DIGITS) 
		{
			apt_string_assign(&vp_mrcp_message->body, s_recog_digits_string, vp_mrcp_message->pool);
		}
		else if (vp_mc_session->role == MC_RECOG_NUMBER) 
		{
			apt_string_assign(&vp_mrcp_message->body, s_recog_number_string, vp_mrcp_message->pool);
		}
	}
	return vp_mrcp_message;
}

/** 
Creates and returns an MRCP RECOGNIZE request message.  
The generic header is generated by the UniMRCP framework and some additions are done in this method: 
-Confidence-Threshold: 0.01
-N-Best-List-Length: 1
-No-Input-Timeout: 15000
-Recognition-Timeout: 10000
-Start-Input-Timers: true
-Cancel-If-Queue: false
-Content-Type:text/uri-list.
-Content = session:request1@form-level.store

Example:
MRCP/2.0 155 RECOGNIZE 543260
Channel-Identifier:32AECB23433801@speechrecog
Confidence-Threshold: 0.01							>>>Added in this method (from here and below) 
N-Best-List-Length: 1
No-Input-Timeout: 15000
Recognition-Timeout: 10000
Start-Input-Timers: true
Cancel-If-Queue: false
Content-Type:text/uri-list
Content-Length:...									     >>>Length value generated by UniMRCP framework

session:request1@form-level.store				 >>>Content

@PARAMS:
 - ap_session : UniMRCP session relevant for the message
 - ap_channel : UniMRCP channel through which the audio stream to be recognized will be sent
 - return     : a fully qualified UniMRCP RECOGNIZE request message 
*/
static mrcp_message_t* app_recognize_message_create(mrcp_session_t * ap_session, 
                                                    mrcp_channel_t * ap_channel)
{
	const char vp_content[] = "session:request1@form-level.store";

	/* create MRCP message */
	mrcp_message_t * vp_mrcp_message = mrcp_application_message_create(ap_session, ap_channel, RECOGNIZER_RECOGNIZE);
	if(vp_mrcp_message) 
	{
		mrcp_recog_header_t   * vp_recog_header;
		mrcp_generic_header_t * vp_generic_header;

		//Add "content-type = text/uri-list" to the generic header
		vp_generic_header = mrcp_generic_header_prepare(vp_mrcp_message);
		if(vp_generic_header) 
		{
			apt_string_assign(&vp_generic_header->content_type, "text/uri-list", vp_mrcp_message->pool);
			mrcp_generic_header_property_add(vp_mrcp_message, GENERIC_HEADER_CONTENT_TYPE);
		}

		//Add several parameters to the RECOGNIZE header
		vp_recog_header = mrcp_resource_header_prepare(vp_mrcp_message);
		if(vp_recog_header) 
		{
			if(vp_mrcp_message->start_line.version == MRCP_VERSION_2) 
			{
				//Set recognizer header fields
				vp_recog_header->cancel_if_queue = FALSE;
				mrcp_resource_header_property_add(vp_mrcp_message, RECOGNIZER_HEADER_CANCEL_IF_QUEUE);
			}
			vp_recog_header->n_best_list_length	= 1;
			mrcp_resource_header_property_add(vp_mrcp_message, RECOGNIZER_HEADER_N_BEST_LIST_LENGTH);
			
			vp_recog_header->no_input_timeout		= 15000;
			mrcp_resource_header_property_add(vp_mrcp_message, RECOGNIZER_HEADER_NO_INPUT_TIMEOUT);
			
			vp_recog_header->recognition_timeout	= 10000;
			mrcp_resource_header_property_add(vp_mrcp_message, RECOGNIZER_HEADER_RECOGNITION_TIMEOUT);
			
			vp_recog_header->start_input_timers	= TRUE;
			mrcp_resource_header_property_add(vp_mrcp_message, RECOGNIZER_HEADER_START_INPUT_TIMERS);
			
			vp_recog_header->confidence_threshold	= 0.01f;
			mrcp_resource_header_property_add(vp_mrcp_message, RECOGNIZER_HEADER_CONFIDENCE_THRESHOLD);
		}

		//Add "session:request1@form-level.store" to link to the grammar defined earlier
		apt_string_assign(&vp_mrcp_message->body, vp_content, vp_mrcp_message->pool);
	}
	return vp_mrcp_message;
}

/** 
Parse NLSML result to extract the recognized digits or number.

Example message:
<result x-model="http://theYesNoModel"
 xmlns:xf="http://www.w3.org/2000/xforms"
 grammar="http://theYesNoGrammar>
  <interpretation>
    <xf:instance>
      <myApp:yes_no>
        <response>yes</response>
      </myApp:yes_no>
    </xf:instance>
    <input>ok</input>
  </interpretation>
</result>

@PARAMS:
 - ap_message : RECOGNITION result message to parse
 - return   	: a string representing the list of recognized digits or the recognized number 
	              e.g. one seven seven six or one thousand...
*/
static char * mc_application_parse_result(mrcp_message_t * ap_message)
{
	apr_xml_elem * vp_interpret;
	apr_xml_elem * vp_instance;
	apr_xml_elem * vp_input;
	char         * vp_result_iteration   = NULL;
	char         * vp_result_aggregation = NULL;

	apr_xml_doc  * vp_doc = nlsml_doc_load(&ap_message->body, ap_message->pool);
	if(!vp_doc) 
	{
		return FALSE;
	}

	//Walk through interpreted results 
	vp_interpret = nlsml_first_interpret_get(vp_doc);
	for(; vp_interpret; vp_interpret = nlsml_next_interpret_get(vp_interpret)) 
	{

		//Get instance and input 
		nlsml_interpret_results_get(vp_interpret, &vp_instance, &vp_input);

		if(vp_instance) 
		{
			//Process instance
			if(vp_instance->first_cdata.first) 
			{
				vp_result_iteration = apr_psprintf(ap_message->pool, "Interpreted Instance [%s]", vp_instance->first_cdata.first->text);
				if (vp_result_aggregation) 
				{
					vp_result_aggregation = apr_psprintf(ap_message->pool, "%s\n%s", vp_result_aggregation, vp_result_iteration);
				}
				else 
				{
					vp_result_aggregation = apr_psprintf(ap_message->pool, "%s", vp_result_iteration);
				}
				apt_log(APT_LOG_MARK, APT_PRIO_INFO, vp_result_iteration);
			}
		}

		if(vp_input) 
		{
			//Process input
			if(vp_input->first_cdata.first) 
			{
				vp_result_iteration = apr_psprintf(ap_message->pool, "Interpreted Input [%s]", vp_input->first_cdata.first->text);
				if (vp_result_aggregation) 
				{
					vp_result_aggregation = apr_psprintf(ap_message->pool, "%s\n%s", vp_result_aggregation, vp_result_iteration);
				}
				else 
				{
					vp_result_aggregation = apr_psprintf(ap_message->pool, "%s", vp_result_iteration);
				}
				apt_log(APT_LOG_MARK, APT_PRIO_INFO, vp_result_iteration);
			}
		}
	}

	return vp_result_aggregation;
}



/** MPF CALLBACKS  **/

/** 
Callback method called from MPF engine context to destroy any additional data associated with audio stream.
Add processing if required.

@PARAMS:
 - ap_stream : MPF audio stream description and codec information
 - return    : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t synth_stream_destroy(mpf_audio_stream_t * ap_stream)
{
	return TRUE;
}

/** 
Callback method called from MPF engine context to perform application stream specific action before open.
Add processing if required.

@PARAMS:
 - ap_stream : MPF audio stream description and codec information
 - ap_codec  : MPF codec attached to the stream (ap_stream) 
 - return    : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t synth_stream_open(mpf_audio_stream_t * ap_stream, 
                                    mpf_codec_t        * ap_codec)
{
	return TRUE;
}

/** 
Callback method called from MPF engine context to perform application stream specific action after close.
Add processing if required.

@PARAMS:
	- ap_stream: audio stream description and codec information
*/
static apt_bool_t synth_stream_close(mpf_audio_stream_t * ap_stream)
{
	return TRUE;
}

/** 
Callback method called from MPF engine context to provide audio data to this client coming
from the MRCP server as a result of a SPEAK request.  The calls are done frame by frame.

@PARAMS:
 - ap_stream : MPF audio stream description and codec information
 - ap_frame  : MPF audio frame to push into a local file
 - return    : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t synth_stream_write(mpf_audio_stream_t * ap_stream, 
                                     const mpf_frame_t  * ap_frame)
{
	mc_app_channel_t * vp_mc_channel = ap_stream->obj;
	if(vp_mc_channel && vp_mc_channel->p_synth_output_file) 
	{
		fwrite(ap_frame->codec_frame.buffer, 1, ap_frame->codec_frame.size, vp_mc_channel->p_synth_output_file);
	}
	return TRUE;
}

/** 
Callback method called from MPF engine context to destroy any additional data associated with audio stream.
Add processing if required.

@PARAMS:
 - ap_stream : MPF audio stream description and codec information
 - return    : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t recog_stream_destroy(mpf_audio_stream_t * ap_stream)
{
	return TRUE;
}

/** 
Callback method called from MPF engine context to perform application stream specific action before open.
Add processing if required.

@PARAMS:
 - ap_stream : MPF audio stream description and codec information
 - ap_codec  : MPF codec description for an RTP channel 
 - return    : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t recog_stream_open(mpf_audio_stream_t * ap_stream, 
                                    mpf_codec_t        * ap_codec)
{
	return TRUE;
}

/** 
Callback method called from MPF engine context to perform application stream specific action after close.
Add processing if required.

@PARAMS:
 - ap_stream : MPF audio stream description and codec information
 - return    : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t recog_stream_close(mpf_audio_stream_t * ap_stream)
{
	return TRUE;
}

/** 
Callback method called from MPF engine context to read a new frame as input to send to the MRCP server 
for recognitions. From the server perspective, this is the source for the inbound RTP stream.
The channel variable "streaming" is turned to TRUE or FALSE from the different callback methods
according to the state or last message received.  If it is on (TRUE), then a frame of the audio file mapped
to channel->p_synth_output_file is added to ap_freame otherwise, nothing is done.  
if channel->p_synth_output_file is NULL, then the frame is filled with silence.

@PARAMS:
 - ap_stream : MPF audio stream description and codec information
 - ap_frame  : MPF audio frame to fill from a local file in order to send data through an RTP channel  
 - return    : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t recog_stream_read(mpf_audio_stream_t * ap_stream, 
                                    mpf_frame_t        * ap_frame)
{
	mc_app_channel_t * vp_mc_channel = ap_stream->obj;

	if(vp_mc_channel && vp_mc_channel->streaming == TRUE) 
	{
		if(vp_mc_channel->p_synth_output_file) 
		{
			//Use the audio file corresponding to "p_synth_output_file" mapped to the channel to send audio to the MRCP server
			//Read/add one frame: will be called multiple times by UniMRCPserver
			if(fread(ap_frame->codec_frame.buffer, 1, ap_frame->codec_frame.size, vp_mc_channel->p_synth_output_file) == ap_frame->codec_frame.size) 
			{
				ap_frame->type |= MEDIA_FRAME_TYPE_AUDIO;
			}
			else 
			{
				//File is over
				vp_mc_channel->streaming = FALSE;
			}
		}
		else 
		{
			//Fill with silence in case no file is available
			if(vp_mc_channel->time_to_complete >= CODEC_FRAME_TIME_BASE) 
			{
				ap_frame->type |= MEDIA_FRAME_TYPE_AUDIO;
				memset(ap_frame->codec_frame.buffer, 0, ap_frame->codec_frame.size);
				vp_mc_channel->time_to_complete -= CODEC_FRAME_TIME_BASE;
			}
			else 
			{
				vp_mc_channel->streaming = FALSE;
			}
		}
	}
	return TRUE;
}


//Client initialization and utility methods------------------------------------

/**
Prints how to call the main executable and the options available.
*/
static void usage()
{
	printf(	"\n"
			"Usage:\n"
			"\n"
			"  unimrcpclient [options]\n"
			"\n"
			"  Available options:\n"
			"\n"
			"	-a number				 : Number of applications to generate \n"
			"   -i number				 : Number if iterations to perform for each application \n"
			"   -r [--root-dir] path     : Set the project root directory path.\n"
			"   -l [--log-prio] priority : Set the log priority.\n"
			"                             (0-emergency, ..., 7-debug)\n"
			"   -o [--log-output] mode   : Set the log output mode.\n"
			"                             (0-none, 1-console only, 2-file only, 3-both)\n"
			"   -h [--help]              : Show the help.\n"
			"\n");
}

/**
Extract and return the command line options.  
If one option is invalid, the list of valid options is printed and "False" is returned.  
Otherwise, the provided options are filled and the function call returns TRUE.

@PARAMS:
 - ap_options : client options to set from command line arguments
 - ap_argc    : number of command line arguments
 - ap_argv    : arguments from the command line
 - return     : TRUE if the execution is successful and FALSE otherwise
*/
static const apt_bool_t mc_util_extract_cmd_arguments(client_options_t   * ap_options, 
                                                      int                  ap_argc, 
                                                      const char * const * ap_argv)
{
	apr_status_t   v_option_status;
	apr_getopt_t * vp_options = NULL;
	int            v_option_id;
	const char   * vp_option_value;

	const apr_getopt_option_t vp_option_meta[] = 
	{
		/* long-option, short-option, has-arg flag, description */
 		{ "num-apps",    'a', TRUE,  "number of apps running in parallel" }, /* -a arg or --num-aps arg */
 		{ "num-iter",    'i', TRUE,  "number of iterations per app" },       /* -i arg or --num-ter arg */
		{ "root-dir",    'r', TRUE,  "path to root dir" },                   /* -r arg or --root-dir arg */
		{ "log-prio",    'l', TRUE,  "log priority" },                       /* -l arg or --log-prio arg */
		{ "log-output",  'o', TRUE,  "log output mode" },                    /* -o arg or --log-output arg */
		{ "help",        'h', FALSE, "show help" },                          /* -h or -#include "apt_pool.h"-help */
		{ NULL, 0, 0, NULL },                                                /* end */
	};

	//Initialize and fill the APR option processing instance with the command line arguments 
	v_option_status = apr_getopt_init(&vp_options, sp_app_pool , ap_argc, ap_argv);
	if(v_option_status != APR_SUCCESS) 
	{
		return FALSE;
	}

	//Process each option/value pair
	while((v_option_status = apr_getopt_long(vp_options, vp_option_meta, &v_option_id, &vp_option_value)) == APR_SUCCESS) 
	{
		switch(v_option_id) 
		{
			case 'a':
				if(vp_option_value) 
				{
					ap_options->num_apps = atoi(vp_option_value);
				}
				break;
			case 'i':
				if(vp_option_value) 
				{
					ap_options->num_iter = atoi(vp_option_value);
				}
				break;
			case 'r':
				ap_options->p_root_dir_path = vp_option_value;
				break;
			case 'l':
				if(vp_option_value) 
				{
					ap_options->log_priority = atoi(vp_option_value);
				}
				break;
			case 'o':
				if(vp_option_value)
				{
					ap_options->log_output = atoi(vp_option_value);
				}
				break;
			case 'h':
				usage();
				return FALSE;
		}
	}

	if(v_option_status != APR_EOF) 
	{
		//Invalid usage, print the correct usage and signal an error
		usage();
		return FALSE;
	}

	return TRUE;       
}

/**
Initialize the APR framework.

@PARAMS:
 - return : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_client_init_apr( )
{
	//APR global initialization
	if(apr_initialize() != APR_SUCCESS) 
	{
		apr_terminate();
		return FALSE;
	}

	//Create APR pool
	sp_app_pool = apt_pool_create();
	if(!sp_app_pool) 
	{
		apr_terminate();
		return FALSE;
	}

	return TRUE;
}

/**
Initializes client options with default values and from command line arguments.

@PARAMS:
 - ap_options : client options to initialize and set from command line arguments
 - ap_argc	  : number of command line arguments
 - ap_argv	  : arguments from the command line
 - return     : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_client_load_options(client_options_t    * ap_options, 
                                         int                   ap_argc, 
                                          const char * const * ap_argv )
{
	//Default values
	ap_options->num_apps          = 1;
	ap_options->num_iter          = 1;
	ap_options->p_root_dir_path   = "../";
	ap_options->log_priority      = APT_PRIO_INFO;
	ap_options->log_output        = LOG_OUTPUT_CONSOLE_FILE;

	//Set from command line arguments
	if(mc_util_extract_cmd_arguments(ap_options, ap_argc, ap_argv) != TRUE) 
	{
		//Command line arguments are invalid, clean-up and signal the problem
		apr_pool_destroy(sp_app_pool);
		apr_terminate();
		return FALSE;
	}

	return TRUE;
}

/**
Initialize the client application by:
1) Initializing APR
2) Loading command line arguments and initializing options
3) Setting global variable values
4) Loading static configuration (done inside UniMRCP framework)
5) Initializing APT logging

@PARAMS
 - ap_argc: number of command line arguments
 - ap_argv: arguments from the command line
 - return : a UniMRCP client initialized from default values, command line options 
            and static configuration.  In case an error occurs, NULL is returned.
*/
static mrcp_client_t * mc_client_init (int                  ap_argc, 
                                       const char * const * ap_argv)
{
	client_options_t   v_options;
	apt_dir_layout_t * vp_dir_layout = NULL;

	if (!mc_client_init_apr( ))
	{
		return NULL;
	}	

	if (!mc_client_load_options(&v_options, ap_argc, ap_argv))
	{
		return NULL;
	}
	
	//Set number of applications, and number of iterations per application
	s_num_apps            = v_options.num_apps;
	s_num_app_iterations  = v_options.num_iter;

	//Create the structure of default directories layout
	vp_dir_layout = apt_default_dir_layout_create(v_options.p_root_dir_path, sp_app_pool);

	//Initialize APT log
	apt_log_instance_create(v_options.log_output, v_options.log_priority, sp_app_pool);

  //@TODO check this condition...
	if((v_options.log_output & APT_LOG_OUTPUT_FILE) == APT_LOG_OUTPUT_FILE)
	{
		//Open the log file
		apt_log_file_open(vp_dir_layout->log_dir_path,"unimrcpclient",MAX_LOG_FILE_SIZE,MAX_LOG_FILE_COUNT,sp_app_pool);
	}

	return unimrcp_client_create(vp_dir_layout);
}

/**
Initializes and registers global mutex and condition variables for apr thread handling.

@PARAMS:
 - return : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_client_init_thread_ctrl( )
{
	//Create mutex for output file writing
	if (apr_thread_mutex_create(&sp_recog_output_mutex, APR_THREAD_MUTEX_NESTED, sp_app_pool) != APR_SUCCESS) 
	{
		apr_terminate();
		return FALSE;
	}

	//Create mutex for application termination counting
	if (apr_thread_mutex_create(&sp_app_complete_mutex, APR_THREAD_MUTEX_UNNESTED, sp_app_pool) != APR_SUCCESS) 
	{
		apr_terminate();
		return FALSE;
	}

	//Create thread condition for client termination
	if (apr_thread_cond_create(&sp_app_complete_cond, sp_app_pool) != APR_SUCCESS) 
	{
		apr_terminate();
		return FALSE;
	}

	return TRUE;
}

/**
Shuts down the MRCP client, apt loggin and apr framework.

@PARAMS:
 - ap_client : UniMRCP client to shut down
 - return    : TRUE 
 */
static apt_bool_t mc_client_stop(mrcp_client_t * ap_client)
{
	mrcp_client_shutdown(ap_client);
	mrcp_client_destroy(ap_client);

	apt_log_instance_destroy();

	apr_pool_destroy(sp_app_pool);
	apr_terminate();

	return TRUE;
}

/**
Creates applications, links them to the client and starts the client.
As a result, the applications are registered with the client and will be started.

@PARAMS:
 - ap_client : UniMRCP client to start and to which the created applications will be attached
 - return    : TRUE if the execution is successful and FALSE otherwise
*/
static apt_bool_t mc_client_application_start(mrcp_client_t * ap_client)
{
	int                 v_appIndex        = 0;
	mc_application_t	* vp_app_registered = apr_palloc(sp_app_pool, sizeof(mc_application_t*) * s_num_apps);
	
	//Create the applications
	for (v_appIndex=0; v_appIndex < s_num_apps; v_appIndex++)
	{
	  const char * vp_name = apr_psprintf(sp_app_pool, "mc%d", v_appIndex);
		
    vp_app_registered[v_appIndex].p_name     = vp_name;
  	vp_app_registered[v_appIndex].p_mrcp_app = mrcp_application_create( mc_application_handler,
                                                                        NULL,
                                                                        sp_app_pool);
		
		mrcp_client_application_register(ap_client, vp_app_registered[v_appIndex].p_mrcp_app, vp_name);
	}

	//Start client and initialize the applications: the client will start processing the applications when ready
	mrcp_client_start(ap_client);
	for(v_appIndex = 0; v_appIndex<s_num_apps; v_appIndex++)
	{
		if (!mc_application_initialize(vp_app_registered[v_appIndex].p_mrcp_app))
		{
			printf("Call to mrcp_application_run failed for application <%s> n", vp_app_registered[v_appIndex].p_name);
			return FALSE;
		}
	}
	return TRUE;
}


//MAIN---------------------------------------------------------------------------

int main(int argc, const char * const *argv)
{
	mrcp_client_t * vp_client = mc_client_init(argc, argv);
	
  //This calls is required in order to have rand() generate different values in each call.
  //Otherwise, the same value is produced each call.
  srand(time(NULL)); 

	//Initialize and start client and applications
	if (!vp_client						          ||
      !mc_client_init_thread_ctrl( )	||
      !mc_client_application_start(vp_client))
	{
		return 0;
	}

	//Wait for the "application complete signal" from each application
	do 
	{
		//This call blocks until apr_thread_cond_signal(sp_app_complete_cond) is called from 
		//mc_application_on_session_terminate in which s_num_apps_completed is incremented.
		apr_thread_cond_wait (sp_app_complete_cond, sp_app_complete_mutex);
	}
	while (s_num_apps_completed < s_num_apps);

	mc_client_stop(vp_client);

	return 0;
}
